#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <sys/time.h>
#include <utime.h>

// Ezt mirt nem szedi ki a dlfcn.h-bl???
#define RTLD_NEXT ((void *) -1l)

#include "inc1"
static int (*next___open)(const char *, mode_t, ...);
static int (*next___open64)(const char *, mode_t, ...);
static int (*next_open)(const char *, mode_t, ...);
static int (*next_open64)(const char *, mode_t, ...);
static int (*next_chdir)(const char *);
static int (*next_fchdir)(int);
static char * (*next_getcwd)(char *, size_t);
static int (*next_execve)(const char *path, char *const argv[], char *const envp[]);

struct mountpoint {
	char *rpath;
	char *vpath;
	struct mountpoint *prev;
	struct mountpoint *next;
};

struct mounttable {
	struct mountpoint *first;
	struct mountpoint *last;
};

struct path {
	char path[PATH_MAX];
	int is_mountpoint;
};

static char *table;
static struct mounttable *mounttable;

static char rcwd[PATH_MAX];
static char vcwd[PATH_MAX];

static int do_act = 0;
static int debuglevel = 0;
#define dprintf(d, x...) if (debuglevel >= d) fprintf(stderr, ##x)

/*
 *  Az els sztringben adott tvonal a msodiknak prefixe-e.
 *  Nagyjbl egy sztring prefix ellenrzs, de vigyzat:
 *  /a/b nem prefixe az /a/bc-nek!
 */
static int is_path_prefix (const char *sht, const char *lng)
{
	int sl, ll;
	sl = strlen(sht);
	ll = strlen(lng);
	if (sl > ll) return 0;
	if (strncmp(sht, lng, sl)) return 0;
	if (sl >= 1 && lng[sl - 1] == '/') return 1;
	if (lng[sl] == '/' || lng[sl] == 0) return 1;
	return 0;
}

/*
 *  Az adott virtulis tvonalhoz megkeresi a hozz tartoz mount pointot.
 */
static struct mountpoint *vget_mountpoint (const char *vpath)
{
	struct mountpoint *mp;
	mp = mounttable->last;
	while (mp != NULL) {
//printf("%s %s %s\n", vpath, mp->vpath, mp->rpath);
		if (is_path_prefix(mp->vpath, vpath)) return mp;
		mp = mp->prev;
	}
	fprintf(stderr, "Internal error in vget_mountpoint()\n");
	exit(1);
}

/*
 *  Az adott igazi tvonalhoz megkeresi a hozz tartoz mount pointot.
 *  NULL-t ad vissza, ha nincs tallat.
 */
static struct mountpoint *rget_mountpoint (const char *rpath)
{
	struct mountpoint *mp;
	mp = mounttable->last;
	while (mp != NULL) {
//printf("%s %s %s\n", rpath, mp->vpath, mp->rpath);
		if (is_path_prefix(mp->rpath, rpath)) return mp;
		mp = mp->prev;
	}
	return NULL;
}

#if 0
/*
 *  mount point-e az adott virtulis tvonal?
 */
static int is_mountpoint (const char *vpath)
{
	struct mountpoint *mp;
	mp = vget_mountpoint(vpath);
	return !strcmp(mp->vpath, vpath);
}
#endif

/*
 *  Virtulis tvonalbl igazit csinl.
 *  Szmt r, hogy a kapott virtulis tvonal kanonikus, abszolt tvonal.
 *  (Gyk. kanonikus = amit a readlink() vissza tud adni, teht nincs benne
 *  szimlink, ., .., dupla / jel.)
 */
static int v_to_r (const char *vpath, char *rpath)
{
	struct mountpoint *mp;
	int i;
	mp = vget_mountpoint(vpath);
	if (mp == NULL) return -1;
	strcpy(rpath, mp->rpath);
	i = strlen(mp->vpath);
	if (vpath[i] == 0) return 0;
	if (rpath[strlen(rpath) - 1] != '/' && vpath[i] != '/')
	  strcat(rpath, "/");
	else if (rpath[strlen(rpath) - 1] == '/' && vpath[i] == '/')
	  i++;
	strcat(rpath, &vpath[i]);
	return 0;
}

/*
 *  Igazi tvonalbl virtulisat csinl.
 *  Szmt r, hogy a kapott igazi tvonal kanonikus, abszolt tvonal.
 */
static int r_to_v (const char *rpath, char *vpath)
{
	struct mountpoint *mp;
	int i;
	mp = rget_mountpoint(rpath);
	if (mp == NULL) return -1;
	strcpy(vpath, mp->vpath);
	i = strlen(mp->rpath);
	if (rpath[i] == 0) return 0;
	if (vpath[strlen(vpath) - 1] != '/' && rpath[i] != '/')
	  strcat(vpath, "/");
	else if (vpath[strlen(vpath) - 1] == '/' && rpath[i] == '/')
	  i++;
	strcat(vpath, &rpath[i]);
	return 0;
}


/*
 *  Adott a kezd virtulis tvonal (vstart) s az innen megteend virtulis
 *  tvonal (vgohere). Vajon hov jutunk el? Az eredmny virtulis tvonalat
 *  megtudjuk vres-bl.
 */
static int _go_here (const char *vstart, const char *vgohere, char *vres,
  int l, int level)
{
	int i;
	char rpath[PATH_MAX];
	char vcur[PATH_MAX];
	char vstep[PATH_MAX];
	char *p;
	struct stat st;

	if (level > 5) {
		errno = ELOOP;
		return -1;
	}
	if (vgohere[0] == 0) {
		errno = ENOENT;
		return -1;
	}
	if (vgohere[0] == '/') {
		dprintf(2, "##         %*s%d %s + %s\n", 2*level, "", level, vstart, vgohere);
		strcpy(vres, "/");
		while (vgohere[0] == '/') vgohere++;

	} else {
		strcpy(vres, vstart);
	}
	while (vgohere[0] != 0) {
		while (vgohere[0] == '/') vgohere++;

		dprintf(2, "##         %*s%d %s + %s\n", 2*level, "", level, vres, vgohere);

		if (vgohere[0] == '.' && (vgohere[1] == '/' || vgohere[1] == 0)) {
			vgohere++;
			while (vgohere[0] == '/') vgohere++;
			continue;
		}
		if (vgohere[0] == '.' && vgohere[1] == '.' &&
		  (vgohere[2] == '/' || vgohere[2] == 0)) {
			vgohere += 2;
			while (vgohere[0] == '/') vgohere++;

			if (!strcmp(vres, "/")) continue;
			p = vres + strlen(vres) - 1;
			while (*p != '/') p--;
			if (p < vres) { fprintf(stderr, "Bazze\n"); exit(1); }
			if (p == vres) p++;
			*(p++) = 0;
			continue;
		}
		i = 0;
		while (vgohere[0] != '/' && vgohere[0] != 0) {
			vstep[i++] = vgohere[0];
			vgohere++;
		}
		vstep[i] = 0;
		strcpy(vcur, vres);
		if (vcur[strlen(vcur) - 1] != '/') {
			vcur[strlen(vcur) + 1] = 0;
			vcur[strlen(vcur)] = '/';
		}
		strcat(vcur, vstep);
		v_to_r(vcur, rpath);
		if ((!l || vgohere[0] != 0) && lstat(rpath, &st) >= 0 && S_ISLNK(st.st_mode)) {
			char buf[PATH_MAX];
			char cur[PATH_MAX];
			i = readlink(rpath, buf, PATH_MAX - 1);
			if (i < 0) return -1;
			buf[i] = 0;
			strcpy(cur, vres);
			if (_go_here(cur, buf, vres, 0, level + 1) < 0) return -1;
		} else {
			strcpy(vres, vcur);
		}
	}
	dprintf(2, "##         %*s%d %s\n", 2*level, "", level, vres);
	return 0;
}

static int go_here (const char *vgohere, char *vres, int l)
{
	return _go_here(vcwd, vgohere, vres, l, 0);
}

static int vpath_to_rpath (const char *vpath, char *rpath, int l)
{
	char vpath2[PATH_MAX];
	go_here(vpath, vpath2, l);
	v_to_r(vpath2, rpath);
//printf("%s -> %s -> %s\n", vpath, vpath2, rpath);
	return 0;
}


void virtualmount_init (void) __attribute((constructor));
void virtualmount_init (void)
{
	int i;
	int rb, re, vb, ve;
	struct mountpoint *mp;
	char *p;

	#include "inc2"
	next___open   = dlsym(RTLD_NEXT, "__open");
	next___open64 = dlsym(RTLD_NEXT, "__open64");
	next_open     = dlsym(RTLD_NEXT, "open");
	next_open64   = dlsym(RTLD_NEXT, "open64");
	next_chdir    = dlsym(RTLD_NEXT, "chdir");
	next_fchdir   = dlsym(RTLD_NEXT, "fchdir");
	next_getcwd   = dlsym(RTLD_NEXT, "getcwd");
	next_execve   = dlsym(RTLD_NEXT, "execve");

	mounttable = malloc(sizeof(struct mounttable));
	mounttable->first = mounttable->last = NULL;
	table = getenv("VIRTUALMOUNT");
	if (table != NULL && table[0] != 0) {
		i = 0;
		while (1) {
			while (table[i] == ' ' || table[i] == '\t' || table[i] == '\n')
				i++;
			if (table[i] == 0) break;
			rb = i;
			if (table[i] != '/') {
				fprintf(stderr, "Path is not absolute!\n");
				exit(1);
			}
			while (table[i] != ' ' && table[i] != '\t' &&
			  table[i] != '\n' && table[i] != '=' && table[i] != 0)
				i++;
			re = i;
			if (table[i] != '=') {
				fprintf(stderr, "Missing = sign!\n");
				exit(1);
			}
			i++;
			vb = i;
			if (table[i] != '/') {
				fprintf(stderr, "Path is not absolute!\n");
				exit(1);
			}
			while (table[i] != ' ' && table[i] != '\t' && table[i] != '\n' &&
			  table[i] != 0)
				i++;
			ve = i;

			mp = malloc(sizeof(struct mountpoint));
			mp->prev = mounttable->last;
			mp->next = NULL;
			if (mounttable->first == NULL) {
				mounttable->first = mp;
			} else {
				mounttable->last->next = mp;
			}
			mounttable->last = mp;

			mp->rpath = malloc(re - rb + 1);
			mp->vpath = malloc(ve - vb + 1);
			strncpy(mp->rpath, &table[rb], re - rb);
			strncpy(mp->vpath, &table[vb], ve - vb);
			mp->rpath[re - rb] = 0;
			mp->vpath[ve - vb] = 0;
		}
	}
	if (mounttable->first == NULL ||
	  strcmp(mounttable->first->vpath, "/")) {
		mp = malloc(sizeof(struct mountpoint));
		mp->next = mounttable->first;
		mp->prev = NULL;
		if (mounttable->last == NULL) {
			mounttable->last = mp;
		} else {
			mounttable->first->prev = mp;
		}
		mounttable->first = mp;

		mp->rpath = "/";
		mp->vpath = "/";
	}

	p = getenv("VIRTUALMOUNT_DEBUG");
	if (p != NULL) debuglevel = atoi(p);

	if (getcwd(rcwd, PATH_MAX) < 0) {
		fprintf(stderr, "getcwd() failed: %s\n", sys_errlist[errno]);
		exit(1);
	}
	if (r_to_v(rcwd, vcwd) < 0) {
		strcpy(vcwd, "/");
		v_to_r(vcwd, rcwd);
		chdir(rcwd);
		dprintf(1, "## Initial working directory changed!\n");
	}
	dprintf(1, "## Initial working directory\n");
	dprintf(1, "##     virt = %s\n", vcwd);
	dprintf(1, "##     real = %s\n", rcwd);

	do_act = 1;
}




#include "inc3"

int __open (const char *path, int flags, ...)
{
	char rpath[PATH_MAX];
	int retval;
	mode_t mode;
	va_list arg;
	va_start(arg, flags);
	if (flags & O_CREAT) {
		mode = va_arg(arg, mode_t);
	}
	va_end(arg);
	if (!do_act) {
		if (flags & O_CREAT) {
			return next___open(path, flags, mode);
		} else {
			return next___open(path, flags);
		}
	}
	do_act = 0;
	dprintf(1, "## __open\n");
	vpath_to_rpath(path, rpath, 0);
	dprintf(1, "##     virt = %s\n", path);
	dprintf(1, "##     real = %s\n", rpath);
	if (flags & O_CREAT) {
		retval = next___open(rpath, flags, mode);
	} else {
		retval = next___open(rpath, flags);
	}
	do_act = 1;
	return retval;
}

int __open64 (const char *path, int flags, ...)
{
	char rpath[PATH_MAX];
	int retval;
	mode_t mode;
	va_list arg;
	va_start(arg, flags);
	if (flags & O_CREAT) {
		mode = va_arg(arg, mode_t);
	}
	va_end(arg);
	if (!do_act) {
		if (flags & O_CREAT) {
			return next___open64(path, flags, mode);
		} else {
			return next___open64(path, flags);
		}
	}
	do_act = 0;
	dprintf(1, "## __open64\n");
	vpath_to_rpath(path, rpath, 0);
	dprintf(1, "##     virt = %s\n", path);
	dprintf(1, "##     real = %s\n", rpath);
	if (flags & O_CREAT) {
		retval = next___open64(rpath, flags, mode);
	} else {
		retval = next___open64(rpath, flags);
	}
	do_act = 1;
	return retval;
}

int open (const char *path, int flags, ...)
{
	char rpath[PATH_MAX];
	int retval;
	mode_t mode;
	va_list arg;
	va_start(arg, flags);
	if (flags & O_CREAT) {
		mode = va_arg(arg, mode_t);
	}
	va_end(arg);
	if (!do_act) {
		if (flags & O_CREAT) {
			return next_open(path, flags, mode);
		} else {
			return next_open(path, flags);
		}
	}
	do_act = 0;
	dprintf(1, "## open\n");
	vpath_to_rpath(path, rpath, 0);
	dprintf(1, "##     virt = %s\n", path);
	dprintf(1, "##     real = %s\n", rpath);
	if (flags & O_CREAT) {
		retval = next_open(rpath, flags, mode);
	} else {
		retval = next_open(rpath, flags);
	}
	do_act = 1;
	return retval;
}

int open64 (const char *path, int flags, ...)
{
	char rpath[PATH_MAX];
	int retval;
	mode_t mode;
	va_list arg;
	va_start(arg, flags);
	if (flags & O_CREAT) {
		mode = va_arg(arg, mode_t);
	}
	va_end(arg);
	if (!do_act) {
		if (flags & O_CREAT) {
			return next_open64(path, flags, mode);
		} else {
			return next_open64(path, flags);
		}
	}
	do_act = 0;
	dprintf(1, "## open64\n");
	vpath_to_rpath(path, rpath, 0);
	dprintf(1, "##     virt = %s\n", path);
	dprintf(1, "##     real = %s\n", rpath);
	if (flags & O_CREAT) {
		retval = next_open64(rpath, flags, mode);
	} else {
		retval = next_open64(rpath, flags);
	}
	do_act = 1;
	return retval;
}

int chdir (const char *path)
{
	char vpath[PATH_MAX];
	char rpath[PATH_MAX];
	int retval;
	if (!do_act) return next_chdir(path);
	do_act = 0;
	go_here(path, vpath, 0);
	v_to_r(vpath, rpath);
	dprintf(1, "## chdir\n");
	dprintf(1, "##     virt = %s\n", path);
	dprintf(1, "##     real = %s\n", rpath);
	retval = next_chdir(rpath);
	if (retval == 0) {
		strcpy(vcwd, vpath);
		strcpy(rcwd, rpath);
	}
	do_act = 1;
	return retval;
}

int fchdir (int fd)
{
	int retval;
	if (!do_act) return next_fchdir(fd);
	do_act = 0;
	retval = next_fchdir(fd);
	getcwd(rcwd, PATH_MAX);
	r_to_v(rcwd, vcwd);
	dprintf(1, "## fchdir\n");
	dprintf(1, "##     virt = %s\n", vcwd);
	dprintf(1, "##     real = %s\n", rcwd);
	do_act = 1;
	return 0;
}

char *getcwd (char *buf, size_t size)
{
	char *retval;
	if (!do_act) return next_getcwd(buf, size);
	do_act = 0;
	dprintf(1, "## getcwd\n");
	dprintf(1, "##     virt = %s\n", vcwd);
	dprintf(1, "##     real = %s\n", rcwd);
	if (buf != NULL) {
		if (size <= strlen(vcwd)) {
			errno = ERANGE;
			retval = NULL;
		} else {
			strcpy(buf, vcwd);
			retval = buf;
		}
	} else {
		if (size == 0) {
			buf = strdup(vcwd);
			retval = buf;
		} else if (size <= strlen(vcwd)) {
			errno = ERANGE;
			retval = NULL;
		} else {
			buf = malloc(size);
			strcpy(buf, vcwd);
			retval = buf;
		}
	}
	do_act = 1;
	return retval;
}

int execve (const char *path, char *const argv[], char *const envp[])
{
	char rpath[PATH_MAX];
	int retval;
	int fd;
	int i;
	int b, e;
	int script = 1;
	int intp_has_arg;
	char intp[128];
	char intpnam[128];
	char intparg[128];
	int argc;
	char **newargv;
	if (!do_act) return next_execve(path, argv, envp);
	do_act = 0;
	dprintf(1, "## execve\n");
	vpath_to_rpath(path, rpath, 0);
	dprintf(1, "##     virt = %s\n", path);
	dprintf(1, "##     real = %s\n", rpath);

	fd = open(rpath, O_RDONLY);
	if (fd < 0) script = 0;
	if (script) {
		e = read(fd, intp, 127);
		close(fd);
		if (e < 2) script = 0;
		intp[e] = 0;
		if (strncmp(intp, "#!", 2)) script = 0;
	}
	if (script) {
		i = 2;
		while (intp[i] != '\n' && intp[i] != 0) i++;
		if (intp[i] != '\n') script = 0;
	}
	if (script) {
		e = i - 1;
		while (e >= 2 && (intp[e] == ' ' || intp[e] == '\t')) e--;
		if (e < 2) script = 0;
	}
	if (script) {
		intp[++e] = 0;
		b = 2;
		while (intp[b] == ' ' || intp[b] == '\t') b++;
		i = b;
		while (intp[i] != ' ' && intp[i] != '\t' && intp[i] != 0) i++;
		intp_has_arg = (intp[i] != 0);
		strncpy(intpnam, &intp[b], i - b);
		intpnam[i - b] = 0;
		if (intp_has_arg) {
			while (intp[i] == ' ' || intp[i] == '\t') i++;
			strncpy(intparg, &intp[i], e - i);
			intparg[e - i] = 0;
		}
		argc = 0;
		while (argv[argc] != NULL) argc++;
		newargv = malloc(sizeof(char *) * (argc + 3));
		for (i = 0; i <= argc; i++) {
			newargv[i + 1 + intp_has_arg] = argv[i];
		}
		if (intp_has_arg) newargv[1] = intparg;
		newargv[0] = intpnam;

		vpath_to_rpath(intpnam, rpath, 0);
		dprintf(1, "##     exec type = script\n");
		dprintf(1, "##         interpreter = %s\n", intpnam);
		if (intp_has_arg) { dprintf(1, "##         interpreter arg = %s\n", intparg); }
		else { dprintf(1, "##         no interpreter arg\n"); }
		dprintf(1, "##     virt = %s\n", intpnam);
		dprintf(1, "##     real = %s\n", rpath);

		retval = next_execve(rpath, newargv, envp);

		free(newargv);
	} else {
		dprintf(1, "##     exec type = binary\n");
		retval = next_execve(rpath, argv, envp);
	}
	do_act = 1;
	return retval;
}

